from __future__ import annotations
import logging
from logging import (
    FileHandler,
    Formatter,
    getLogger,
    Logger,
    StreamHandler,
)
from pathlib import Path
from typing import Optional, TextIO
from typing_extensions import Literal

__all__ = [
    "get_logger",
]


def get_logger(
        name: Optional[str] = None,
        level: int = logging.INFO,
        propagate: Optional[bool] = None,
        stream: Optional[TextIO] = None,
        file_path: Optional[str | Path] = None,
        file_mode: str = "a",
        encoding: Optional[str] = None,
        format_str: Optional[str] = None,
        date_format: Optional[str] = None,
        format_style: Literal["%", "{", "$"] = "%"
) -> Logger:
    """Create or retrieve logger based on name and configurations

    Args:
        name: logger name
        level: log level
        propagate: if propagate to parent loggers
        stream: logging stream. If not None, will remove all existing handlers.
        file_path: logging file path. If not None, will remove all existing handlers.
        file_mode: logging file mode.
        encoding: logging file encoding.
        format_str: format string. If not None, will be registered to all handlers.
        date_format: date format string. If not None, will be registered to all handlers.
        format_style: formatter style.

    Returns:
        logger

    """
    logger = getLogger(name)
    logger.setLevel(level)

    handlers = []

    if isinstance(propagate, bool):
        logger.propagate = propagate

    if stream is not None:
        handlers.append(StreamHandler(stream))

    if file_path is not None:
        handlers.append(FileHandler(filename=str(file_path), mode=file_mode, encoding=encoding))

    if len(handlers) > 0:
        # remove old handlers
        if logger.hasHandlers():
            for h in list(logger.handlers):
                logger.removeHandler(h)

        # register new handlers
        for h in handlers:
            logger.addHandler(h)

    if format_str is not None or date_format is not None:
        formatter = Formatter(format_str, datefmt=date_format, style=format_style)

        for h in list(logger.handlers):
            h.setFormatter(formatter)

    return logger



from __future__ import annotations
from functools import wraps
import os
from typing import (
    Any,
    Callable,
    Optional,
    TypeVar,
)
from typing_extensions import Literal, ParamSpec

from torch import distributed, Tensor
from torch.distributed import ProcessGroup, ReduceOp

T = TypeVar("T")
DT = TypeVar("DT")
P = ParamSpec("P")

GroupType = Optional[ProcessGroup]
ReduceOpStrType = Literal["mean", "average", "sum"]

__all__ = [
    "all_reduce",
    "barrier",
    "get_local_rank",
    "get_rank",
    "get_world_size",
    "is_distributed",
    "one_rank_only",
    "safe_distributed",
]


def is_distributed() -> bool:
    return distributed.is_available() and distributed.is_initialized()


def safe_distributed(default: DT = None) -> Callable[P, T | DT]:
    """
    Decorator that returns default value when not in distributed mode

    Args:
        default: default value

    Returns:

    """

    def _safe_distributed(func: Callable[P, T]) -> Callable[P, T | DT]:
        @wraps(func)
        def wrapped_func(*args: P.args, **kwargs: P.kwargs) -> T | DT:
            if is_distributed():
                return func(*args, **kwargs)
            else:
                return default

        return wrapped_func

    return _safe_distributed


@safe_distributed(default=None)
def barrier() -> Optional[Any]:
    return distributed.barrier()


@safe_distributed(default=1)
def get_world_size(group: GroupType = None) -> int:
    return distributed.get_world_size(group)


@safe_distributed(default=0)
def get_local_rank() -> int:
    return os.environ.get("LOCAL_RANK", 0)


@safe_distributed(default=0)
def get_rank(group: GroupType = None) -> int:
    return distributed.get_rank(group)


def get_reduce_operation():
    pass


def all_reduce(
        tensor: Tensor,
        op: ReduceOp | Literal["mean", "average", "sum"] = ReduceOp.SUM,
        group: GroupType = None,
        inplace: bool = False
) -> Tensor:
    outputs = tensor if inplace else tensor.clone()

    if is_distributed():
        distributed.all_reduce(outputs, op=op, group=group)

    return outputs


def one_rank_only(rank: int = 0, synchronize: bool = False) -> Callable[P, Optional[T]]:
    """

    Args:
        rank: target rank to execute the function.
        synchronize: if True, will synchronize with a barrier.

    Returns:

    """

    def _one_rank_only(func: Callable[P, T]) -> Callable[P, Optional[T]]:
        @wraps(func)
        def wrapped_func(*args: P.args, **kwargs: P.kwargs) -> Optional[T]:
            output = None
            try:
                if get_rank() == rank:
                    output = func(*args, **kwargs)
            finally:
                if synchronize:
                    barrier()
            return output

        return wrapped_func

    return _one_rank_only



from typing import Iterable, Optional

from pandas import DataFrame

from ..utils import get_progress_bar

__all__ = [
    "merge_dataframes",
]


def merge_dataframes(dataframes: Iterable[DataFrame]) -> Optional[DataFrame]:
    output_df: Optional[DataFrame] = None

    for df in get_progress_bar(dataframes, desc="Merging"):
        if output_df is None:
            output_df = df
        else:
            output_df = output_df.merge(
                df,
                how="outer",
                on=output_df.columns.intersection(df.columns).tolist(),
                validate="1:1"
            )

    return output_df



from __future__ import annotations
from typing import Optional, Sequence, Union

import numpy as np
import torch
from torch import Size, Tensor

TensorInputType = Union[Sequence[Union[int, float]], np.ndarray, Tensor]

__all__ = [
    "random_choice",
    "to_tensor",
]


def to_tensor(inputs: int | float | TensorInputType) -> Tensor:
    if isinstance(inputs, Tensor):
        return inputs
    elif isinstance(inputs, np.ndarray):
        return torch.from_numpy(inputs)
    else:
        return torch.tensor(inputs)


def random_choice(
        choices: TensorInputType,
        size: Size | Sequence[int],
        replacement: bool = True,
        weights: Optional[TensorInputType] = None,
        dtype: Optional[torch.dtype] = None,
        device: Optional[torch.device] = None
) -> Tensor:
    """
    See https://discuss.pytorch.org/t/torch-equivalent-of-numpy-random-choice/16146/14

    Args:
        choices: Candidates to draw from.
        size: Output shape.
        replacement: Whether to draw with replacement or not.
        weights: The probabilities associated with each candidate.
        dtype: Output data type.
        device: Device of output data.

    Returns:

    """
    choices = to_tensor(choices).to(dtype=dtype, device=device)
    if len(choices.shape) != 1:
        raise ValueError("choices must be a 1D tensor/numpy array/list/tuple")

    if weights is None:
        weights = torch.ones_like(choices, dtype=torch.float)
    else:
        weights = to_tensor(weights).to(dtype=torch.float, device=device)
    if len(weights.shape) != 1:
        raise ValueError("weights must be a 1D tensor/numpy array/list/tuple")

    if weights.shape != choices.shape:
        raise RuntimeError("choices and weights must have the same shape")

    indices = torch.multinomial(
        weights,
        num_samples=np.prod(size).item(),
        replacement=replacement
    )

    outputs: Tensor = torch.index_select(choices, 0, indices)

    return outputs.reshape(size).contiguous()



import re
from typing import Mapping

__all__ = [
    "replace",
]


def replace(text: str, replacements: Mapping[str, str]) -> str:
    # see https://stackoverflow.com/a/6117124
    pattern = re.compile("|".join(re.escape(k) for k in replacements))
    return pattern.sub(lambda m: replacements.get(m.group(0), m.group(0)), text)



from __future__ import annotations
import json
from json import JSONDecodeError, JSONEncoder
from pathlib import Path
from typing import (
    Any,
    AnyStr,
    IO,
    Optional,
)

from .path import is_file

__all__ = [
    "DataEncoder",
    "safe_load",
    "safe_loads",
    "is_json",
]


class DataEncoder(JSONEncoder):
    def default(self, obj: Any) -> Any:
        if isinstance(obj, set):
            return list(obj)

        return super().default(obj)


def safe_load(file: IO[AnyStr], default: Optional[Any] = None, **kwargs: Any) -> Optional[Any]:
    try:
        outputs = json.load(file, **kwargs)
    except JSONDecodeError:
        outputs = default

    return outputs


def safe_loads(string: AnyStr, default: Optional[Any] = None, **kwargs: Any) -> Optional[Any]:
    try:
        outputs = json.loads(string, **kwargs)
    except JSONDecodeError:
        outputs = default

    return outputs


def is_json(file_path: str | Path, mode: str = "r", encoding: Optional[str] = "utf-8", **kwargs: Any) -> bool:
    if not is_file(file_path):
        return False

    try:
        with open(file_path, mode=mode, encoding=encoding, **kwargs) as f:
            json.load(f)
    except JSONDecodeError:
        return False
    else:
        return True



from __future__ import annotations
import itertools
import time
from functools import wraps
from typing import (
    Any,
    Callable,
    Iterable,
    Optional,
    Sequence,
)

from tqdm import tqdm

from .typing import ExceptionType, verify_exception_type

__all__ = [
    "flatten",
    "get_progress_bar",
    "retry",
    "sleep",
    "split_chunks",
    "unwrap_tqdm",
]


def split_chunks(inputs: Sequence[Any], chunk_size: int) -> list[list[Any]]:
    assert chunk_size > 0
    return [list(inputs[i:i + chunk_size]) for i in range(0, len(inputs), chunk_size)]


def retry(
        max_retry: int = 0,
        backoff_factor: float = 0,
        exception_type: ExceptionType = Exception
) -> Callable[..., Any]:
    assert max_retry >= 0
    assert verify_exception_type(exception_type)

    def decorate(func: Callable[..., Any]) -> Callable[..., Any]:
        @wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            for i in range(max_retry + 1):
                try:
                    result = func(*args, **kwargs)
                    return result
                except exception_type as e:
                    if i >= max_retry:
                        raise e

                    else:
                        sleep(backoff_factor)

                    # continue the loop
                    continue

        return wrapper

    return decorate


def flatten(lists: Iterable[Iterable[Any]]) -> list[Any]:
    return list(itertools.chain.from_iterable(lists))


def unwrap_tqdm(iterable: Iterable | tqdm) -> Optional[Iterable]:
    if isinstance(iterable, tqdm):
        iterable = iterable.iterable

    return iterable


def get_progress_bar(
        iterable: Iterable,
        is_tqdm: bool = True,
        desc: Optional[str] = None,
        total: Optional[int | float] = None,
        **kwargs: Any
) -> tqdm | Optional[Iterable]:
    if "disable" in kwargs:
        raise ValueError("is_tqdm is mutually exclusive with disable")

    if not is_tqdm:
        return unwrap_tqdm(iterable)

    elif not isinstance(iterable, tqdm):
        # if is_tqdm and iterable is not tqdm
        return tqdm(iterable, desc=desc, total=total, **kwargs)

    else:
        return iterable


def sleep(seconds: float) -> None:
    """
    Sleep for `seconds` seconds if `seconds` > 0 else do nothing

    Args:
        seconds: number of seconds to sleep

    Returns:

    """
    if seconds > 0:
        time.sleep(seconds)



from __future__ import annotations
from pathlib import Path
from typing import (
    Any,
    Optional,
    Sequence,
    Union,
)

import cv2 as cv
import numpy as np
from PIL import Image, ImageDraw, ImageFont
from PIL.Image import Image as PILImage

from .path import is_file

ImageType = Union[PILImage, np.ndarray]

__all__ = [
    "cv_to_pil",
    "draw_text",
    "pil_open",
    "pil_to_cv",
    "thumbnail",
    "to_cv",
    "to_pil",
]


def pil_open(filename: str | Path, color_mode: Optional[str] = None) -> Image:
    image = Image.open(str(filename))

    if color_mode is not None:
        image = image.convert(color_mode)

    return image


def cv_to_pil(image: np.ndarray) -> PILImage:
    return Image.fromarray(cv.cvtColor(image, cv.COLOR_BGR2RGB))


def pil_to_cv(image: PILImage) -> np.ndarray:
    return cv.cvtColor(np.asarray(image.convert("RGB")), cv.COLOR_RGB2BGR)


def to_pil(image: ImageType, inplace: bool = False) -> PILImage:
    """
    Convert image to PIL.Image.Image.

    Args:
        image: input image
        inplace: if True, will return the same image if no conversion is needed; else, return a copy

    Returns:

    """
    if isinstance(image, PILImage):
        if inplace:
            return image
        else:
            return image.copy()

    elif isinstance(image, np.ndarray):
        return cv_to_pil(image)

    else:
        raise NotImplementedError(f"{type(image)} is not a supported image type")


def to_cv(image: ImageType, inplace: bool = False) -> np.ndarray:
    """
    Convert image to np.ndarray.

    Args:
        image: input image
        inplace: if True, will return the same image if no conversion is needed; else, return a copy

    Returns:

    """
    if isinstance(image, PILImage):
        return pil_to_cv(image)

    elif isinstance(image, np.ndarray):
        if inplace:
            return image
        else:
            return image.copy()

    else:
        raise NotImplementedError(f"{type(image)} is not a supported image type")


def thumbnail(image: ImageType, max_size: int | float, inplace: bool = False) -> PILImage:
    """
    Limit the image size to be at most (max_size, max_size) while keeping the original ratio

    Args:
        image: input image
        max_size: max size
        inplace: if True, will resize the original image; else, return a copy

    Returns:

    """
    image = to_pil(image, inplace=inplace)
    if not np.isinf(max_size):
        image.thumbnail(size=(max_size, max_size))

    return image


def draw_text(
        image: ImageType,
        text: str,
        coordinates: tuple[int, int],
        font_path: Optional[str | Path] = None,
        font_size: int = 32,
        anchor: Optional[str] = None,
        color: Sequence[int] | str = "red",
        background_color: Optional[Sequence[int] | str] = None,
        **kwargs: Any
) -> ImageType:
    """

    Args:
        image: input image
        text: text to draw
        coordinates: (x, y) coordinates of the input image
        font_path: .ttf or .ttc file
        font_size: font size
        anchor: The text anchor alignment. See https://pillow.readthedocs.io/en/stable/handbook/text-anchors.html#text-anchors
        color: color string or RGB color tuple
        background_color: color string or RGB color tuple for background. If None, no background

    Returns:

    """
    assert font_path is None or is_file(font_path), f"\"{font_path}\" is not a valid font file path"
    output_image = to_pil(image)

    draw = ImageDraw.Draw(output_image)

    # see https://stackoverflow.com/a/50854463
    font = ImageFont.truetype(font_path, font_size) if font_path is not None else draw.getfont()

    if background_color is not None:
        # see https://stackoverflow.com/a/18869779
        x, y = coordinates
        box = font.getbbox(text, anchor=anchor)

        # draw background
        draw.rectangle((box[0] + x, box[1] + y, box[2] + x, box[3] + y), fill=background_color)

    draw.text(
        coordinates,
        text,
        font=font,
        fill=color,
        anchor=anchor,
        **kwargs
    )

    if isinstance(image, np.ndarray):
        output_image = to_cv(output_image)

    return output_image



from __future__ import annotations
from contextlib import ContextDecorator, contextmanager
from typing import Any, Dict, ContextManager

from torch.nn import Module
from torch.nn.parallel import DataParallel, DistributedDataParallel

__all__ = [
    "consume_prefix_in_state_dict_if_present",
    "get_module_size",
    "set_model_mode",
    "unwrap_ddp",
]


@contextmanager
def set_model_mode(model: Module, mode: bool) -> ContextManager[None] | ContextDecorator:
    """

    Args:
        model: input model
        mode: True turns on set training mode; False turns on evaluation mode

    Returns:

    """
    prev_mode = model.training
    model.train(mode=mode)

    try:
        yield
    finally:
        model.train(prev_mode)


def consume_prefix_in_state_dict_if_present(
        state_dict: Dict[str, Any],
        prefix: str
) -> None:
    r"""Copied from https://github.com/pytorch/pytorch/blob/57f039b41f940af4f02718fdf967cca1f713d759/torch/nn/modules/utils.py#L43

    Strip the prefix in state_dict in place, if any.
    ..note::
        Given a `state_dict` from a DP/DDP model, a local model can load it by applying
        `consume_prefix_in_state_dict_if_present(state_dict, "module.")` before calling
        :meth:`torch.nn.Module.load_state_dict`.
    Args:
        state_dict (OrderedDict): a state-dict to be loaded to the model.
        prefix (str): prefix.
    """
    keys = sorted(state_dict.keys())
    for key in keys:
        if key.startswith(prefix):
            newkey = key[len(prefix):]
            state_dict[newkey] = state_dict.pop(key)

    # also strip the prefix in metadata if any.
    if "_metadata" in state_dict:
        metadata = state_dict["_metadata"]
        for key in list(metadata.keys()):
            # for the metadata dict, the key can be:
            # '': for the DDP module, which we want to remove.
            # 'module': for the actual model.
            # 'module.xx.xx': for the rest.

            if len(key) == 0:
                continue
            newkey = key[len(prefix):]
            metadata[newkey] = metadata.pop(key)


def get_module_size(module: Module) -> int:
    return sum(p.numel() for p in module.parameters())


def unwrap_ddp(module: Module) -> Module:
    if isinstance(module, (DataParallel, DistributedDataParallel)):
        return module.module
    else:
        return module



from typing import (
    Any,
    Tuple,
    Type,
    Union,
)

ExceptionType = Union[Type[Exception], Tuple[Type[Exception], ...]]

__all__ = [
    # types
    "ExceptionType",

    # functions
    "verify_exception_type",
]


def verify_exception_type(exception_type: Any) -> bool:
    """
    Check if exception_type is Type[Exception] | Tuple[Type[Exception], ...]

    Args:
        exception_type: a type object

    Returns:

    """
    if isinstance(exception_type, type):
        return issubclass(exception_type, Exception)

    elif isinstance(exception_type, tuple):
        return all(isinstance(t, type) and issubclass(t, Exception) for t in exception_type)

    else:
        return False



from __future__ import annotations
import math
from typing import Any, Callable, Optional

__all__ = [
    "get_file_type_str",
    "get_optional_file_type_str",
    "get_optional_type",
    "int_or_inf",
    "optional_str",
]


def int_or_inf(input_str: str) -> int | float:
    result = float(input_str)

    if not math.isinf(result):
        result = int(input_str)

    return result


def get_optional_type(func: Callable[[str], Any]) -> Callable[[str], Optional[Any]]:
    def wrapper(input_str: str) -> Optional[Any]:
        if input_str.lower().strip() in {"none", "null"}:
            return None
        else:
            return func(input_str)

    return wrapper


def optional_str(input_str: str) -> Optional[str]:
    return get_optional_type(lambda x: x)(input_str)


def get_file_type_str(*suffix: str) -> Callable[[str], str]:
    suffix = tuple(s.lower().strip() for s in suffix)
    assert all(s.startswith(".") for s in suffix), "suffix must begin with \".\""

    def func(input_str: str) -> str:
        input_str = input_str.strip()
        assert input_str.lower().endswith(suffix), f"\"{input_str}\" does not end with \"{suffix}\""
        return input_str

    return func


def get_optional_file_type_str(*suffix: str) -> Callable[[str], Optional[str]]:
    return get_optional_type(get_file_type_str(*suffix))



"""
References:
    - https://setuptools.pypa.io/en/latest/userguide/quickstart.html
    - https://packaging.python.org/tutorials/packaging-projects/
"""
from pathlib import Path
from pkg_resources import parse_requirements
import platform
import re
from setuptools import setup, find_packages
import sys

NAME = "lomltk"
DESCRIPTION = "The lord of Machine Learning toolkits"
MIN_PYTHON_VERSION = (3, 7, 0)

if sys.version_info < MIN_PYTHON_VERSION:
    print(f"You are using Python {platform.python_version()}. Python >= {MIN_PYTHON_VERSION} is required.")
    sys.exit(-1)


def read(*paths, **kwargs):
    kwargs.setdefault("encoding", "utf8")

    with Path(*paths).open(**kwargs) as f:
        return f.read()


def read_requirements(*paths, **kwargs):
    kwargs.setdefault("encoding", "utf8")

    with Path(*paths).open(**kwargs) as f:
        return [str(req) for req in parse_requirements(f)]


def read_version(*paths, **kwargs):
    version_file = read(*paths, **kwargs)
    match = re.search(r"__version__ = ['\"]([^'\"]*)['\"]", version_file)

    if match is not None:
        return match.group(1)
    else:
        return "UNKNOWN"


# Setting up
setup(
    name=NAME,
    version=read_version("lomltk/__init__.py"),
    author="Zijian Hu",
    author_email="zijian-hu@outlook.com",
    description=DESCRIPTION,
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    url="https://github.com/zijian-hu/lomltk",
    packages=find_packages(include=["lomltk"]),
    include_package_data=True,
    install_requires=read_requirements("requirements.txt"),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.7"
)



from __future__ import annotations
import os
from pathlib import Path
from typing import Callable

from joblib import delayed, Parallel

from .utils import get_progress_bar
from .multiprocessing import resolve_num_workers, tqdm_joblib_context

__all__ = [
    "get_real_path",
    "is_file",
    "is_dir",
    "find_all_files_helper",
    "find_all_files",
    "safe_delete",
]


def get_real_path(file_path: str | Path) -> Path:
    return Path(os.path.realpath(str(file_path)))


def is_file(file_path: str | Path) -> bool:
    return get_real_path(file_path).is_file()


def is_dir(file_path: str | Path) -> bool:
    return get_real_path(file_path).is_dir()


def find_all_files_helper(
        input_dir: str | Path,
        is_valid_func: Callable[[str | Path], bool] = is_file,
        pattern: str = "*",
        is_recursive: bool = True,
        num_workers: int = 1
) -> tuple[list[Path], list[bool]]:
    input_dir = Path(input_dir)
    num_workers = resolve_num_workers(num_workers)

    file_paths: list[Path] = [
        p for p in get_progress_bar(
            input_dir.rglob(pattern) if is_recursive else input_dir.glob(pattern),
            desc=f"Finding files w/ pattern \"{pattern}\""
        ) if is_file(p)
    ]

    with tqdm_joblib_context(
            get_progress_bar(file_paths, desc="Verifying files")
    ) as progress_bar:
        is_valid_list = Parallel(n_jobs=num_workers)(
            delayed(is_valid_func)(file_path)
            for file_path in progress_bar
        )

    assert len(file_paths) == len(is_valid_list)
    return file_paths, is_valid_list


def find_all_files(
        input_dir: str | Path,
        helper_func: Callable[..., tuple[list[Path], list[bool]]] = find_all_files_helper,
        is_valid_func: Callable[[str | Path], bool] = is_file,
        pattern: str = "*",
        is_recursive: bool = True,
        num_workers: int = 1
) -> list[Path]:
    file_paths, is_valid_list = helper_func(
        input_dir=input_dir,
        is_valid_func=is_valid_func,
        pattern=pattern,
        is_recursive=is_recursive,
        num_workers=num_workers
    )

    return [file_path for file_path, is_valid in zip(file_paths, is_valid_list) if is_valid]


def safe_delete(path: str | Path) -> None:
    try:
        os.remove(str(path))
    except FileNotFoundError:
        return



from __future__ import annotations
from contextlib import ContextDecorator, contextmanager
import os
from typing import ContextManager

import joblib
from joblib import Parallel
from tqdm import tqdm

__all__ = [
    "cpu_count",
    "resolve_num_workers",
    "tqdm_joblib_context",
]


def cpu_count(only_physical_cores: bool = False) -> int:
    """
    Get number of CPU cores on the current machine. If environment variable `MY_CPU_LIMIT` is set,
        will choose the lower value between `joblib.cpu_count` and `$MY_CPU_LIMIT`

    Args:
        only_physical_cores:

    Returns:

    """
    num_cpus = joblib.cpu_count(only_physical_cores)

    return min(num_cpus, int(os.environ.get("MY_CPU_LIMIT", num_cpus)))


def resolve_num_workers(num_workers: int = -1) -> int:
    if num_workers >= 0:
        return max(1, min(num_workers, cpu_count()))

    elif num_workers == -1:
        # use all cpus
        return cpu_count()

    else:
        # num_workers < -1
        return num_workers + 1 + cpu_count() if abs(num_workers) < cpu_count() else 1


@contextmanager
def tqdm_joblib_context(tqdm_object: tqdm) -> ContextManager[tqdm] | ContextDecorator:
    """
    Context manager to patch joblib to report into tqdm progress bar given as argument.
    See https://stackoverflow.com/a/61689175 and https://stackoverflow.com/a/58936697 for detail

    Args:
        tqdm_object: tqdm object

    Returns:

    """

    def tqdm_print_progress(self: Parallel) -> None:
        if self.n_completed_tasks > tqdm_object.n:
            n_completed = self.n_completed_tasks - tqdm_object.n
            tqdm_object.update(n=n_completed)

    old_print_progress = joblib.parallel.Parallel.print_progress
    joblib.parallel.Parallel.print_progress = tqdm_print_progress

    try:
        yield tqdm_object
    finally:
        joblib.parallel.Parallel.print_progress = old_print_progress
        tqdm_object.close()



from typing import Dict, Union

import torch
from torch import Tensor
import numpy as np
import random

RandomState = Dict[str, Union[Tensor, tuple]]

__all__ = [
    # types
    "RandomState",

    # functions
    "get_random_state",
    "set_random_state",
]


def get_random_state() -> RandomState:
    return dict(
        torch=torch.random.get_rng_state(),
        numpy=np.random.get_state(),
        python=random.getstate(),
    )


def set_random_state(random_state: RandomState) -> None:
    torch.random.set_rng_state(random_state["torch"].cpu())
    np.random.set_state(random_state["numpy"])
    random.setstate(random_state["python"])
